//
//  LariatFragment.h  Defines raw data blocks for 
//  various Lariat readout components (W.Badgett)
//
//  Specific board fragment classes should inherit 
//  from this class
//
#ifndef _LariatFragment_h
#define _LariatFragment_h

#include <iostream>
#include <istream>
#include <ios>
#include <limits>
#include <stdint.h>
#include <vector>
#include <string>
#include "WUTFragment.h"
#include "V1495Fragment.h"
#include "CAENFragment.h"
#include "TDCFragment.h"
#include "LARASICFragment.h"
#include "ReadoutError.h"
#include "TriggerFragment.h"

class LariatFragment
{
 public:

  enum RunTypes
    {
      RUNTYPE_PHYSICS            = 1,
      RUNTYPE_PEDESTAL           = 2,
      RUNTYPE_CHANNEL_SCAN       = 3,
      RUNTYPE_TEST               = 4,
      RUNTYPE_PEDESTAL_EXTERNAL  = 5
    };

  enum LariatFragmentType
    {
      FRAGMENT_TYPE_RUNINFO               = 1,
      FRAGMENT_TYPE_SPILLINFO             = 2,
      FRAGMENT_TYPE_CAEN_ADC              = 3,
      FRAGMENT_TYPE_V1495                 = 4,
      FRAGMENT_TYPE_SPILLTRAILER          = 5,
      FRAGMENT_TYPE_TDC_CONTROLLER_HEADER = 6,
      FRAGMENT_TYPE_TDC_TDC_SPILL_HEADER  = 7,
      FRAGMENT_TYPE_TDC_TDC_EVENT_HEADER  = 8,
      FRAGMENT_TYPE_WUT_HIT_HEADER        = 9,
      FRAGMENT_TYPE_LARIAT_CONFIG         = 10,
      FRAGMENT_TYPE_LARASIC_CONFIG        = 11,
      FRAGMENT_TYPE_READOUT_ERROR         = 12,
      FRAGMENT_TYPE_TRIGGER_FRAGMENT      = 13,
      FRAGMENT_TYPE_N                     = 14
    };

#pragma pack(push,1)
    struct RunInfo 
    {
      uint32_t fragmentSize;
      uint32_t fragmentType;
      uint32_t runNumber;
      uint32_t runType;
      uint32_t recordLength;
      uint32_t nBuffers;
      uint32_t timeStamp;
      uint32_t nBoards;
      uint32_t  link;
      uint32_t  boardId[CAENFragment::MAX_BOARDS];
      uint32_t  postPercent;
      uint32_t  maxEventsPerTransfer;
      uint32_t  eventsPerInterrupt;
      uint32_t  irqWaitTime;
      uint32_t  usePedestals;
      uint32_t  dacValue;
      uint32_t  samplingRate;
      uint32_t  sampleReduction;
      char      tdcHostName[80];
      uint32_t  tdcPort;
      uint32_t  nTDCs;
      uint32_t  tdcEnableMask;
      uint32_t  tdcPulserEnable;
      uint32_t  tdcNBoards;
      uint32_t  tdcTimeOut;
    };

    struct LariatConfig
    {
      uint32_t fragmentSize;
      uint32_t fragmentType;
      char   * xml;
    };

    struct SpillInfo 
    {
      uint32_t fragmentSize;
      uint32_t fragmentType;
      uint32_t runNumber;
      uint32_t spillNumber;
      uint32_t timeStamp;
    };

    struct SpillTrailer
    {
      uint32_t fragmentSize;
      uint32_t fragmentType;
      uint32_t runNumber;
      uint32_t spillNumber;
      uint32_t timeStamp;
      double   spillSnapshot; // In milliseconds, time of MidSpill snapshot
      double   spillOffset;   // In seconds, normally 2.000 (msec least count)
      // To get time of $21, spillTime = spillSnap - ( spillOffset * 1000.0) in milliseconds
      int spillV1740;
      int spillV1751;
      int countV1740;
      int countV1751;
    };

#pragma pack(pop)
    
    static const char* fragmentNameFromType(int id);

    //    /**
    //     * stream the data from this object into a buffer.
    //     * Returns the number of characters written as a convenience.
    //     * This function sets fragmentSize and fragmentType.
    //     **/
    //    uint32_t streamToBuffer(std::ostream& buffer);

    // Constructors
    LariatFragment() {}
    LariatFragment(std::istream & buffer);
    LariatFragment(std::string & fileName);
    LariatFragment(const char * buffer, size_t size);
    /**
     * Fill in this object, assuming that the current position
     * in the buffer coincides with the start of this object.
     * Returns the number of characters read as a convenience.
     * Obviously streamToBuffer() and fillFromBuffer() must agree on
     * the structure of the object in the buffer!
     **/
    uint32_t fillFromBuffer(std::istream& buffer);
    uint32_t fillRunInfoFromBuffer(std::istream& buffer);
    uint32_t fillSpillInfoFromBuffer(std::istream& buffer);
    uint32_t fillSpillTrailerFromBuffer(std::istream& buffer);
    uint32_t fillLariatConfigFromBuffer(std::istream& buffer);
    static void writeReadoutError(std::ostream * output, 
				  uint32_t runNumber,
				  uint32_t spillNumber,
				  std::string message);
    
    RunInfo         runInfo;
    LariatConfig    lariatConfig;
    SpillInfo       spillInfo;    
    SpillTrailer    spillTrailer;
    TriggerFragment triggerFragment;
    std::vector<WUTFragment>          wutFrags;
    std::vector<V1495Fragment>        v1495Frags;
    std::vector<CAENFragment>         caenFrags;
    std::vector<TDCFragment>          tdcFrags;
    std::vector<LARASICFragment>      larasicFrags;
    std::vector<ReadoutError>         errorFrags;

    uint32_t nBytes;
    int count[FRAGMENT_TYPE_N];

    void print(std::ostream & os = std::cout);
    void printSpillTrailer(std::ostream & os = std::cout);
    void printLariatConfig(std::ostream & os = std::cout);

    static void checkOStream(std::ostream &buffer);

};

std::ostream& operator<<(std::ostream& os, const LariatFragment& e);
std::ostream& operator<<(std::ostream& os, const 
			 LariatFragment::LariatConfig& e);
std::ostream& operator<<(std::ostream& os, const LariatFragment::RunInfo& e);
std::ostream& operator<<(std::ostream& os, const LariatFragment::SpillInfo& e);
std::ostream& operator<<(std::ostream& os, 
			 const LariatFragment::SpillTrailer& e);

#endif
